/*
	Program that finds triangles in INPUT pgm image
	Outputs points in triangles.

	Sergio Herrera May/2018
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "matrix.h"
#include "pnm.h"
#include "chamfer.h"

#define CINC 5
#define FILLTONE	25

int do_work( char *namein );
void negate( MATRIX* );
void umbral( MATRIX* , int );
void findOBJ( const MATRIX *IMG , int* X, int* Y , int COLOR);
void findFurthestPx( const MATRIX *IMG ,int X, int Y, int* Fx, int* Fy , int COLOR);
void findFurthestVec( const MATRIX *IMG ,int X1, int Y1, int X2 ,int Y2, int* X3, int* Y3 , int COLOR);
void drawPoint( MATRIX *IMG , int X, int Y , int R , int C);
void cleanBorder( MATRIX *IN, int R);

extern MATRIX * convolution ( MATRIX *IN, MATRIX *FIL, int ka );


int main( int argc, char *argv[] )
{
	char namein[64];

	if ( argc < 2 ) { 
		fprintf( stderr, "  [!]Use: %s input_image.pbm \n", argv[0]);
		return -1;
	}

	strncpy( namein, argv[1], 64 );
	//strncpy( nameout, argv[2], 64 );
	//ope = atoi( argv[3] );
	//U = atoi( argv[4] );
	
	if ( do_work( namein ) >= 0 ) {
		printf("# Done! \n");
		//fprintf( stderr, "# Done!!! \n" );
		return 0;
	}
	else {
		fprintf( stderr, "ERROR: Something is wrong!!!\n" );
		exit(1);
	}

	return 0;
}

int do_work( char *namein ) 
{
	MATRIX *IMG, *IN, *FIL;

	IN = read_pnm( namein ); 

	if ( IN ==NULL ) {
		fprintf( stderr, "ERROR reading %s file\n", namein );
		return -1;
	}
	if ( IN->type != 5 ) {
		fprintf( stderr, "ERROR Bad format. Image must be in b/w\n");
		return -1;
	}
	
	//2 - Smoothing (2 Pass)
		int **fil;
		FIL = matrix_allocate( 3, 3, sizeof( int ) );
		if ( FIL==NULL ) {
			fprintf( stderr, "ERROR allocating memory\n" );
			return -1;
		}
		//Smooth Window
		fil = (int **)FIL->ptr;
		fil[0][0] = 1; fil[0][1] = 1; fil[0][2] = 1;
		fil[1][0] = 1; fil[1][1] = 1; fil[1][2] = 1;
		fil[2][0] = 1; fil[2][1] = 1; fil[2][2] = 1;

		IMG = convolution( IN, FIL, 9 );	//1st Pass
		//write_pnm( IMG, "convol1.pgm" );

		//memcpy( IN, IMG, sizeof(MATRIX) );
		//IMG = convolution( IN, FIL, 9 );	//2nd Pass
		//write_pnm( IMG, "convol2.pgm" );

	
	//3 - Binarize IMG global threshold 
	umbral( IMG, 97 );

	//4 - Opening / Erotion->Dilatation
	int Umbral = 2;
	//Opening
			//Erosion
			negate( IMG );
			chamfer34( IMG ); 
			umbral( IMG , Umbral );
			negate( IMG );
			//write_pnm( IMG, "EROTION.pgm" );
			//Dilatation
			chamfer34( IMG );
			umbral( IMG , Umbral );
			//write_pnm( IMG, "DILATATION.pgm" );
			cleanBorder( IMG , Umbral );
			//write_pnm( IMG, "DILATATION2.pgm" );

	//5 - Extract triangles (Fill)
		printf("#Looking for objects... \n");
		int X,Y;
		int OBJcount = 0;
		
		//Look for objects to fill them
		while(1)
		{
			findOBJ(IMG, &X, &Y, 0);
			if(X > -1){ //Valid Point
				//printf("#Found object at: %d , %d \n", X,Y);
				push(X,Y);
				do_fill(IMG, OBJcount * CINC);
					//drawPoint( IMG , X, Y , 10 , 255);
				OBJcount+=1;
			}
			else break;
		}
		printf("# \t [%d] Objects found! \n", OBJcount);

	//6 - Compute vertices of each triangle
		for (int i = 0; i < OBJcount; ++i)//***
		{
			int COLOR = FILLTONE + i * CINC;
			findOBJ(IMG, &X, &Y, COLOR); //Find first vertex
			if(X > -1){ //Valid Point
				int X2,Y2, X3,Y3;

				printf("#Object [%d] \n",i+1);
				//printf("Found object at: %d , %d \n", X,Y);
					//drawPoint( IMG , X, Y , 10 , 255);
				//printf("Looking for color : %d \n", COLOR);
				findFurthestPx(IMG, X, Y, &X2, &Y2, COLOR); //Find 2nd vertex 
				//printf("Found furthest at: %d , %d \n", X2,Y2);
				findFurthestVec(IMG, X,Y,X2,Y2 , &X3, &Y3, COLOR); //Find 3rd vertex 
				//printf("Found 3rd at: %d , %d \n", X3,Y3);
				printf("\t %d,%d \t %d,%d \t %d,%d \n", X,Y, X2,Y2, X3,Y3);

			}
		}
		//a) Find first pixel of OBJColor

	//7 - Output points ???

	//if ( write_pnm( IMG, "fillTest1.pgm" ) != 0 )  
    //    return -1;

	matrix_free( IMG );
	matrix_free( IN );
	matrix_free( FIL );

	return 0;
}


// --------------------
// 
void drawPoint( MATRIX *IMG , int X, int Y , int R , int C) 
{
	byte *px;
	for (int i = Y-R; i < Y+R; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = X-R; j < X+R; j++){
			*px++ = C;
		}
	}
}

// --------------------
// Finds furthest Pixel given a vector discibed by P2-P1
void findFurthestVec( const MATRIX *IMG ,int X1, int Y1, int X2, int Y2, int* X3, int* Y3 , int COLOR)
{
	byte *px;
	long maxDist = 0;
	long dist = 0;
	int Vx,Vy, Ux,Uy, Wx,Wy;

	Vx = X2-X1;
	Vy = Y2-Y1;
	Ux = -Vy;
	Uy = Vx;

	*X3 = *Y3= -1; //No OBJ found!

	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			if((*px) == COLOR){
				//printf("[%d,%d]= %d  ",i,j,*px);
				Wx = j-X1;
				Wy = i-Y1;
				dist = abs((Ux * Wx) + (Uy * Wy));
				if(dist > maxDist){
					maxDist = dist;
					*Y3=i;
					*X3=j;
				}
			}
			px++;
		}
	}
}


// --------------------
// Finds furthest Pixel given X,Y in IMG
void findFurthestPx( const MATRIX *IMG ,int X, int Y, int* Fx, int* Fy , int COLOR)
{
	byte *px;
	long maxDist = 0;
	long dist = 0;

	*Fy = *Fx= -1; //No OBJ found!

	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			if((*px) == COLOR){
				//printf("[%d,%d]= %d  ",i,j,*px);
				dist = (X-j)*(X-j) + (Y-i)*(Y-i);
				if(dist > maxDist){
					maxDist = dist;
					*Fy=i;
					*Fx=j;
				}
			}
			px++;
		}
	}
}


// --------------------
// Find first OBJ pixel in IMG
void findOBJ( const MATRIX *IMG , int* X, int* Y , int COLOR) 
{
	byte *px;
	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			if((*px) == COLOR){
				//printf("[%d,%d]= %d  ",i,j,*px);
				*Y=i;
				*X=j;
				return;
			}
			px++;
		}
	}
	*Y=	*X= -1; //No OBJ found!
}

// --------------------
// Negates binary image
void negate( MATRIX *IMG )
{
	byte *px;
	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			*px = ((*px) == 255)? 0:255;
			px++;
		}
	}
}

// --------------------
// Clean clean!
void cleanBorder( MATRIX *IN,  int Radius )
{
	int i, j ;
	byte *ptrin;
	/** limpio del borde de la imagen  **/
	
	//Horizontal
    for (int k = 0; k < Radius; ++k)
    {	
    ptrin = (byte *) IN->ptr[0 + k];
    for( j=0; j<IN->cols; j++ ) *ptrin++ = 255;
    
    ptrin = (byte *) IN->ptr[ IN->rows - 1 -k ];
    for( j=0; j<IN->cols; j++ ) *ptrin++ = 255;
    }
	
	//Vertical
	for (int k = 0; k < Radius; ++k)
    {
    for( i=1; i<IN->rows-1; i++ ){
        ptrin = (byte *) IN->ptr[i] + k ;
        *ptrin = 255;
        ptrin = (byte *) IN->ptr[i] + IN->cols - 1 - k ;
        *ptrin = 255;
    }
	}
	
    
}


// --------------------
// Binarizes image given an umbral
void umbral( MATRIX *IMG, int umbral)
{
	byte *px;
	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			*px = ((*px) > umbral)? 255:0;
			px++;
		}
	}
}





